#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <strings.h>

using namespace std;

#include "bitfield.h"


/* initializes the bitfield depending on the number of pieces*/
void bitfield_setup(bool is_seed,unsigned char* bf, unsigned int npieces)
{
	/* REQUIRES */
	assert(npieces <= MAX_PIECES);

	if(is_seed)
	{
		for(int i = 0 ; i < MAX_BITFIELD_LEN; i++)
		{
			bf[i] = 0xFF;
		}

	}
	else
	{

		unsigned int req_chars;
		req_chars = npieces/8;

		bzero(bf,sizeof(unsigned char) * MAX_BITFIELD_LEN);
		for(int i = req_chars ; i < MAX_BITFIELD_LEN; i++)
		{
			bf[i] = 0xFF;
		}


		if(npieces % 8 > 0)
		{
			unsigned char mask1 = 0x7F;
			unsigned char mask2 = 0xFF;

			mask1 = mask1>>((npieces%8) - 1);
			mask2 = mask1 & mask2;
			bf[req_chars] = bf[req_chars] & mask2;
		}
	}	

}

/* checks to see if the other side has the given piece */
bool other_side_has(unsigned int piece, unsigned char* bf)
{
	unsigned int which_char = piece/8;
	/* REQUIRES*/
	assert(which_char < MAX_BITFIELD_LEN);
	
	unsigned int index = piece%8;
	
	unsigned char mask = 0x80;

	mask = mask>>index;

	if(mask & (bf[which_char]))
	{
		return true;
	}
	else
	{
		return false;
	}
}

/* checks to see if the other side has something I don't have */
bool has_something(unsigned char* local_bf, unsigned char* other_bf)
{
	/* if same bitfields => nothing interesting */
	if(memcmp(local_bf, other_bf, MAX_BITFIELD_LEN) == 0)
	{
		return false;
	}

	
	/* not same => check if I don't have something but he does */
	for(int i = 0; i < MAX_BITFIELD_LEN ; i++)
	{
		if(local_bf[i] == other_bf[i])
		{
			continue;
		}
		else
		{	
			unsigned char mask = 0x80;
			for(int j = 0;j < 8; j++)
			{
				if((!(mask & local_bf[i])) && (mask & other_bf[i]))
				{
					return true;
				}
				mask = mask >> 1;
			}
		}
	}

	return false;
}

/* adds a piece to a bitfield => sets the corresponding bit */
void add_to_bitfield(unsigned char* bf, unsigned int piece)
{
	unsigned int which_char = piece/8;
	/* REQUIRES*/
	assert(which_char < MAX_BITFIELD_LEN);
	
	unsigned int index = piece%8;
	
	unsigned char mask = 0x80;

	mask = mask>>index;

	if(mask & (bf[which_char]))
	{
		fprintf(stderr,"Already have the piece! \n");
		return;
	}
	else
	{
		bf[which_char] = bf[which_char] | mask;
	}
}

/* checks that is my bitfield showing all pieces are available or not */
bool i_am_complete(unsigned char* bf)
{
	for(int i = 0; i < MAX_BITFIELD_LEN ; i++)
	{
		if(bf[i] != 0xFF)
			return false;
	}	

	return true;
}

