
#ifndef _BT_DOWNLOADER_H_
#define _BT_DOWNLOADER_H_


class bt_peer;
class uploader;

/* Class to manage the download for a peer => one object per connection*/
class downloader{

public:
	
	/* pointer to peer this downloader object belongs to */
	bt_peer* peer;

	/* pointer to other side of connection */
	uploader* other_side;

	/* bitfield of the pieces other side has */
	unsigned char* other_have;	

	/* time since rate calc shud not have exceeded*/
	unsigned int ratesince_time;

	/* time since last rate calc*/
	unsigned int last_time;

	/*curr estimate of D/L rate */	
	unsigned int dl_rate;


	/* STATE VARIABLES */
	unsigned char choked ;
	unsigned char interested ;

	/* piece being downloaded currently and the amount of it already downloaded*/
	unsigned int curr_piece;
	unsigned int curr_piece_size_downloaded;

	/* constructor and starter functions => called when the connection has been fully setup*/
	downloader(bt_peer* peer);
	void startup();

	/* function to send 'interested'*/
	void send_interested();

	/* function to send 'not interested' */
	void send_not_interested();

	/* function to send 'have' */
	void send_have();

	/* function to receive 'have' */
	void got_have(unsigned int);

	/* function to request another piece*/
	void request_more();

	/* function to receive the data from the network
	 * params: 1) number of bytes received */
	void got_data(unsigned int);

	/* function to receive a piece */
	void got_piece(unsigned int);

	/* function to receive a choke message */
	void got_choke();

	/* function to receive an unchoke message */
	void got_unchoke();

	/* function to update download rate */
	void update_rate(unsigned int amount);


	/* cleans up state associated with the downloader*/
	void cleanup_state();
	/* called up by the other side */
	void close_connection();
	

};

#endif
