#include <iostream>
#include <stdio.h>
#include <vector>
#include <ext/hash_map>
#include <cstdlib>
#include <assert.h>
#include <new>

using namespace std;


#include "bt_tracker.h"
#include "bt_peer.h"
#include "bitfield.h"
#include "topology.h"
#include "event.h"
#include "btsim.h"
#include "util.h"

namespace stdext = ::__gnu_cxx;

int flag_set = 0;


/* constructor for bt_peer */
void bt_peer::join(bool is_seed, unsigned int id, unsigned int npieces)
{
	/* define the time for which this guy remains in the network after it has finished downloading */
	this->grace_seed_time = DEFAULT_SEED_TIME;

	this->peer_id = id;
	this->numpieces = npieces;
	this->local_bitfield = new unsigned char[MAX_BITFIELD_LEN];
	bitfield_setup(is_seed, this->local_bitfield, this->numpieces);

	this->numinterests = new unsigned int[this->numpieces];
	this->pos_in_interests = new unsigned int[this->numpieces];	
	vector<unsigned int>* pieces_0_peers = new vector<unsigned int>();
	
	for(unsigned int i=0; i < this->numpieces; i++)
	{
		this->numinterests[i] = 0;
		this->pos_in_interests[i] = i;
		pieces_0_peers->push_back(i);
	}
	
	this->interests.push_back(pieces_0_peers);

	this->is_active_peer = true;
	if(is_seed)
	{
		this->peer_is_seed = true;
		this->done = true;
		this->numgot = npieces;
	}
	else
	{
		this->peer_is_seed = false;
		this->done = false;
		this->numgot = 0;
	}
	this->rrq_count = 0;
	this->choke_count = 0;
	this->numconnections = 0;

	this->rerequest();

	struct event* choke_event = event_alloc();
	/* FIRE CHOKE EVENT AFTER 30 secs */
	choke_event->fire_time = g_tick + INITIAL_CHOKE_INTERVAL*TICKS_PER_SECOND;
	choke_event->peer = this;
	choke_event->type = E_OPTUNCHOKE;
	event_insert(choke_event);
}

void bt_peer::leave()
{
	/* simply call the close_all_connections function */
	this->is_active_peer = false;
	this->close_all_connections();

	bt_tracker* mytracker = get_tracker(this);
	log_msg("Leave: %u: %u %u\n", g_tick, this->peer_id, mytracker->clients.size());
}

/* this fucntion is called to check if the peer is idle currently
 * ------ used in logging and analyzing performance -----------*/
bool bt_peer::peer_is_idle()
{
	if(this->is_active_peer == false)
		return true;

	/* i am assuming that a peer is idle if 
	 * => none of its downloader is currently downloading
         * => this inactivity maybe because of
         *                          => i am not interested in the pieces the other side
         *                          => i was choked
         *                          => both of the above                     */

	bool rval = true;
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator ul_it_choked;
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator ul_it_interested;
	for(ul_it_choked = this->ul_choked.begin(); ul_it_choked != this->ul_choked.end(); ul_it_choked++)
	{	
		ul_it_interested = this->ul_interested.find(ul_it_choked->first);
		if(ul_it_interested == this->ul_interested.end())
		{
			if(this->done == true)
				fprintf(stderr,"CHECK(DONE)-> Peer: %u, Numgot: %u, Other Side: %u\n", this->peer_id, this->numgot, ul_it_choked->first->peer_id);
			else
				fprintf(stderr,"CHECK(NOT DONE)-> Peer: %u, Numgot: %u, Other Side: %u\n", this->peer_id, this->numgot, ul_it_choked->first->peer_id);

			if(ul_it_choked->first->is_active_peer)
			{
				fprintf(stderr,"Other Side active <->, Numgot: %u\n", ul_it_choked->first->numgot);
			}
			else
			{

				fprintf(stderr,"Other Side not active \n");
			}

		}
		assert(ul_it_interested != this->ul_interested.end());
		/* this means that the interested variable is also found*/

		if (ul_it_interested->second == INTERESTED && ul_it_choked->second == UNCHOKED)
		{
			return false;
		}
	}

	/* this means that the peer is idle*/

	return rval;
}

int bt_peer::num_downloads_curr()
{
	if(this->is_active_peer == false)
		return 0;

	/* i am assuming that a peer is idle if 
	 * => none of its downloader is currently downloading
         * => this inactivity maybe because of
         *                          => i am not interested in the pieces the other side
         *                          => i was choked
         *                          => both of the above                     */

	int rval = 0;
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator dl_it_choked;
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator dl_it_interested;
	for(dl_it_choked = this->dl_choked.begin(); dl_it_choked != this->dl_choked.end(); dl_it_choked++)
	{	
		dl_it_interested = this->dl_interested.find(dl_it_choked->first);
		assert(dl_it_interested != this->dl_interested.end());
		/* this means that the interested variable is also found*/

		if (dl_it_interested->second == INTERESTED && dl_it_choked->second == UNCHOKED)
		{
			rval++;
		}
	}

	return rval;
}





/* this fucntion is called periodically and is used to log whatever I want this peer to log */
void bt_peer::log_peer_info()
{
		//DEBUG_MSG("COMPLETE : peer: %u, numconnections: %u, numpieces %u\n", this->peer_id, this->numconnections, this->numgot);
		//DEBUG_MSG("INCOMPLETE : peer: %u, numconnections: %u, numpieces %u\n", this->peer_id, this->numconnections,this->numgot);
		
}


/* this function is called when this PEER is leaving the network 
 * cleanup PEER state
 * cleanup all downloaders and uploaders state
 * call close connections for other_sides of all downloader and uploaders */
void bt_peer::close_all_connections()
{

	bt_tracker* mytracker = get_tracker(this);
	mytracker->peer_leaving(this);

	/* first of all send a close to other sides => simulataneoulsy clean up the local down/up loaders*/
	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator dl_it;

	for(dl_it = this->dl_choked.begin(); dl_it != this->dl_choked.end(); dl_it++ )
	{
		bt_peer* other_side = dl_it->first;
		/* call the close of the other side downloader/uploader*/
		other_side->download_close_connection(this);
		other_side->upload_close_connection(this);
		other_side->check_min_connections();

		/* cleanup state for this uploader/downloader*/
		download_cleanup_state(other_side,NO_CLEAN_CHOKE);
		upload_cleanup_state(other_side);
	}
	this->dl_choked.clear();

	/* now cleanup the state of the PEER */
	this->choker_connections.clear();
	vector<vector<unsigned int>* >::iterator vvit;
	for(vvit = this->interests.begin(); vvit != this->interests.end(); vvit++)
	{
		vector<unsigned int>* curr_vec = *vvit;
		delete curr_vec;
	}
	this->interests.clear();
	delete [] this->numinterests;
	delete [] this->pos_in_interests;
}


/* function to CONTACT the TRACKER*/
void bt_peer::rerequest()
{

	/* only to be done if this peer is active */
	if(this->is_active_peer)
	{
		/* Get the tracker pointer*/
		bt_tracker* mytracker = get_tracker(this);

		/* get list of peers from client */
		vector<bt_peer*> *got_list = mytracker->announce(this);

		vector<bt_peer*>::iterator peer_it;

			/* traverse the list of peers received*/
			for(peer_it = got_list->begin(); peer_it != got_list->end(); peer_it++)
			{
				if(this->numconnections < MAX_SELF_INITIATE)
				{
					int status = this->start_connection(*peer_it);
					if(status == NEW_PEER)
						this->numconnections++;
				}
				else
				{
					/* Add this peer to spare connections */
					break;
				}
			}

		log_msg("Rerequest: %u: %u: %u: %u: %u: %u: %u\n", g_tick, this->peer_id,got_list->size(), this->numconnections, this->numgot, this->pieces_in_progress.size(), this->num_downloads_curr()) ;

		struct event* rerequest_event = event_alloc();
		/* Next REREQUEST is AFTER 10 MINS. */
		rerequest_event->fire_time = g_tick + REREQUEST_INTERVAL*TICKS_PER_SECOND;
		rerequest_event->peer = this;
		rerequest_event->type = E_REREQUEST;
		event_insert(rerequest_event);
	}

}

/* function to connect to a peer ---
 *  needs: 1) @ param : pointer to the other peer */
int bt_peer::start_connection(bt_peer* other_peer)
{
	/* first check for self connections */
	if(other_peer->peer_id == this->peer_id)
	{
		return EXISTING_PEER;
	}


	stdext::hash_map<bt_peer*, unsigned char, HashBtPeer, HashBtPeerCompare>::iterator dl_it;

	dl_it = this->dl_choked.find(other_peer);

	if( dl_it != this->dl_choked.end())
	{
		return EXISTING_PEER;
	}

	/* check if the other side has maxed the limits*/
	if(other_peer->numconnections >= MAX_CONNECT)
	{
		return MAX_CONNECTED_PEER;
	}

	/* this is a new peer */
	log_msg("NBR: %u %u\n", this->peer_id, other_peer->peer_id);

	/* create a new downloader object*/
	this->dl_choked[other_peer] = CHOKED;
	this->dl_interested[other_peer] = NOT_INTERESTED;
	this->dl_rate[other_peer] = 0;
	this->dl_ratesince_time[other_peer] = 0;
	this->dl_last_time[other_peer] = 0;
	this->curr_piece[other_peer] = NO_NEXT_PIECE;
	this->curr_piece_downloaded[other_peer] = 0;

	/* create a new uploader object */
	this->ul_choked[other_peer] = CHOKED;
	this->ul_interested[other_peer] = NOT_INTERESTED;
	this->ul_rate[other_peer] = 0;
	this->ul_ratesince_time[other_peer] = 0;
	this->ul_last_time[other_peer] = 0;

	/* establish connection with other side */
	other_peer->external_connection_made(this);

	/* also add the uploader to choker connection */
	unsigned int conn_len = this->choker_connections.size();
	int random_int = -2 + (int)( (conn_len+ 1) * rand()/(RAND_MAX + 1.0));
	vector<bt_peer*>::iterator choker_conn_it = this->choker_connections.begin();


	unsigned char foundpeer = 0;

	vector<bt_peer*>::iterator tmpit;
	for(tmpit = this->choker_connections.begin(); tmpit != this->choker_connections.end(); tmpit++)
	{
		if(*tmpit == other_peer)
		{
			fprintf(stderr,"Warning: Found peer already in start_conn %u %u \n", this->peer_id, other_peer->peer_id);
			foundpeer = 1;
		}
		if((*tmpit)->peer_id == other_peer->peer_id)
		{

			fprintf(stderr,"Warning: Found peer already(ID match) in start_conn %u %u \n", this->peer_id, other_peer->peer_id);
			foundpeer = 1;
		}
	}


	if(foundpeer == 0)	
	{
		if(random_int <= 0)
		{
			this->choker_connections.insert(choker_conn_it, other_peer);
		}
		else
		{
			choker_conn_it += random_int;
			this->choker_connections.insert(choker_conn_it, other_peer);
		}
	}
	/* start the downloader*/
	download_startup(other_peer);

	return NEW_PEER;
}

/* connection made by another peer --
 * input: @ pointer to the caller (bt_peer)
          			                    */

void bt_peer::external_connection_made(bt_peer* other_peer)
{
	this->numconnections++;
	stdext::hash_map<bt_peer*, unsigned char,HashBtPeer, HashBtPeerCompare>::iterator dl_it;

	dl_it = this->dl_choked.find(other_peer);

	if( dl_it != this->dl_choked.end())
	{
		return;
	}

	/* this peer doesn't exist */
	/* create a new downloader object*/
	this->dl_choked[other_peer] = CHOKED;
	this->dl_interested[other_peer] = NOT_INTERESTED;
	this->dl_rate[other_peer] = 0;
	this->dl_ratesince_time[other_peer] = 0;
	this->dl_last_time[other_peer] = 0;
	this->curr_piece[other_peer] = NO_NEXT_PIECE;
	this->curr_piece_downloaded[other_peer] = 0;

	/* create a new uploader object */
	this->ul_choked[other_peer] = CHOKED;
	this->ul_interested[other_peer] = NOT_INTERESTED;
	this->ul_rate[other_peer] = 0;
	this->ul_ratesince_time[other_peer] = 0;
	this->ul_last_time[other_peer] = 0;


	/* also add the uploader to choker connection */
	unsigned int conn_len = this->choker_connections.size();
	int random_int = -2 + (int)( (conn_len+ 1) * rand()/(RAND_MAX + 1.0));
	vector<bt_peer*>::iterator choker_conn_it = this->choker_connections.begin();

	
	unsigned char foundpeer = 0;
	vector<bt_peer*>::iterator tmpit;
	for(tmpit = this->choker_connections.begin(); tmpit != this->choker_connections.end(); tmpit++)
	{
		if(*tmpit == other_peer)
		{
			fprintf(stderr,"Warning: Found peer already in ext_conn %u %u \n", this->peer_id, other_peer->peer_id);
			foundpeer = 1;
		}
		if((*tmpit)->peer_id == other_peer->peer_id)
		{

			fprintf(stderr,"Warning: Found peer already(ID match) in ext_conn %u %u \n", this->peer_id, other_peer->peer_id);
			foundpeer = 1;
		}
	}

	if(foundpeer == 0)
	{
		if(random_int <= 0)
		{
			this->choker_connections.insert(choker_conn_it, other_peer);
		}
		else
		{
			choker_conn_it += random_int;
			this->choker_connections.insert(choker_conn_it, other_peer);
		}
	}

	/* start the downloader*/
	download_startup(other_peer);
}

 /*********************************************************************************
 * data sent by the other side
 * this function is called by the topology to notify me of the amount of data 
 * that could have been transferred from the sender to me since last notification 
 * params: 1) @ pointer to the sender peer
 *         2) number of bytes that were received                                 
 **********************************************************************************/
void bt_peer::receive_bytes(bt_peer* other_peer, unsigned int numbytes)
{
	/* only valid for active peers*/
	if(this->is_active_peer)
	{
		if(this->done != true)
		{
			/* find corresponding downloader from the stdext::hash_map */
			stdext::hash_map<bt_peer*, unsigned int,HashBtPeer, HashBtPeerCompare>::iterator dl_it = this->curr_piece_downloaded.find(other_peer);

			if(dl_it != this->curr_piece_downloaded.end())
			{
				/* found the piece info*/
				got_data(other_peer, numbytes);	
			}
			else
			{
				fprintf(stderr,"Protocol Error: Data to/from wrong peer\n");
				return;
			}
		}
	}
}

/***********************************CHOKING MECHANISM*******************************************************
* Maintains   1) Number of maximum uploads possible
*    &        2) List of connections (special list => order of elements is important)
*             3) Whether I am done or not
*  Needs      4) Download/Upload rates from/to all peers.
************************************************************************************************************/
/* function to do choking ---
 * 				*/
void bt_peer::choker()
{
	/*only to be done for active peers */
	if(this->is_active_peer)
	{
		/* if this is a 30-sec interval call => optimistically unchoke */
		if(choke_count++%3 == 0)
		{
			vector<bt_peer*>::iterator ccit;
			vector<bt_peer*>::iterator chosenit;
			for(ccit = this->choker_connections.begin(); ccit != this->choker_connections.end(); ccit++)
			{
				if((this->ul_interested[*ccit] == INTERESTED) && (this->ul_choked[*ccit] == CHOKED))
				{
					chosenit = ccit;
					bt_peer* chosen = *chosenit;
					this->choker_connections.erase(chosenit);
					
					unsigned char foundpeer = 0;
					vector<bt_peer*>::iterator tmpit;
					for(tmpit = this->choker_connections.begin(); tmpit != this->choker_connections.end(); tmpit++)
					{
						if(*tmpit == chosen)
						{
							fprintf(stderr,"Warning: Found peer already in choker %u %u \n", this->peer_id, chosen->peer_id);
							foundpeer = 1;
						}
						if((*tmpit)->peer_id == chosen->peer_id)
						{

							fprintf(stderr,"Warning: Found peer already(ID match) in choker %u %u \n", this->peer_id, chosen->peer_id);
							foundpeer = 1;
						}
					}
					
					if(foundpeer == 0)
					{
						this->choker_connections.insert(this->choker_connections.begin(),chosen);
					}
					break;
				}
			}
		}
		/* if this is a 10-sec interval call => only choke the slowest */
		rechoke();

		struct event* choke_event = event_alloc();
		/* FIRE CHOKE EVENT AFTER 10 secs */
		choke_event->fire_time = g_tick + CHOKE_INTERVAL * TICKS_PER_SECOND;
		choke_event->peer = this;
		choke_event->type = E_OPTUNCHOKE;
		event_insert(choke_event);
	}
}

/* helper function for choker => does the actual decision-making */
void bt_peer::rechoke()
{
	if(this->is_active_peer == false)
	{
		return;
	}


	vector<bt_peer*>* preferred = new vector<bt_peer*>();
	unsigned int max_rate = 0;	
	unsigned int comp_rate;

	map<bt_peer*, bool> rcmap;

	for(unsigned int j = 0; j < this->choker_connections.size() ; j++)
	{
		bt_peer* curr = this->choker_connections.at(j);
		preferred->push_back(curr);
		rcmap[curr] = false;
	}

	int curr_size;
	if(this->choker_connections.size() < MAX_UPLOADS)
		curr_size = this->choker_connections.size();
	else
		curr_size = MAX_UPLOADS;



	for(int i = 0; i < curr_size - 1; i++)
	{
		vector<bt_peer*>::iterator pit;
		vector<bt_peer*>::iterator curr_best_it;
		bt_peer* best_uploader = NULL;

		/* BUG FIXED BY RUCHIR -- max_rate = 0 for every loop */
		max_rate = 0;

		for(pit = preferred->begin(); pit != preferred->end(); pit++)
		{
			bt_peer* curr = *pit;;
			if(this->done)
			{
				comp_rate = this->ul_rate[curr];
			}
			else
			{
				comp_rate = this->dl_rate[curr];
			}

			////DEBUG_MSG("comp_rate %u, max_rate %u\n", comp_rate, max_rate);
	
			if((comp_rate >= max_rate) && (this->ul_interested[curr] == INTERESTED))
			{
				curr_best_it = pit;
				best_uploader = *pit;
				max_rate = comp_rate;
			}
		}
		/* unchoke curr_best*/
		if(best_uploader != NULL)
		{
			if(rcmap[best_uploader] == true)
			{
				fprintf(stderr,"Found peer again in rechoke (best) %u %u\n", this->peer_id, best_uploader->peer_id);
			}
			else
			{
				send_unchoke(best_uploader);
				rcmap[best_uploader] = false;
			}
			preferred->erase(curr_best_it);
		}
	}
	
	bool hit = false;
	vector<bt_peer*>::iterator pit;

	for(pit = preferred->begin(); pit != preferred->end(); pit++)
	{
		bt_peer* curr = *pit;
		if(hit == false)
		{
			if(rcmap[curr] == true)
			{
				fprintf(stderr,"Found peer again in rechoke (random) %u %u\n", this->peer_id, curr->peer_id);
			}
			else
			{
				send_unchoke(curr);
				rcmap[curr] = false;
			}

			if(this->ul_interested[curr] == INTERESTED)
			{
				hit = true;
			}
		}
		else
		{
			if(rcmap[curr] == true)
			{
				fprintf(stderr,"Found peer again in rechoke (choke case random) %u %u\n", this->peer_id, curr->peer_id);
			}
			else
			{
				send_choke(curr);
				rcmap[curr] = false;
			}
		}
	}

	rcmap.clear();

	preferred->clear();
	delete preferred;
}



/* --------------------------PIECE CHOOSING MECHANISM --------------------------------------------------------
* Maintains: 1) A list of how many peers have each piece (numinterests)
*            2) A list of lists such that the i'th list is the list of pieces which are currently with i peers
*     &      3) A list which tells what is the position of a piece in the list in which the piece belongs
*               (pertaining to (2)), which list the piece belongs can be found using (1) and (2)
*   Needs    4) Which pieces have been requested and are currently in midst of transfer
*            5) How many pieces have been completed
-------------------------------------------------------------------------------------------------------------*/

/* function to choose which is the next piece to download*/
unsigned int bt_peer::piece_picker(unsigned char* other_have)
{
	/* if numgot < rarest_first_cutoff => choose a random piece */
	if(this->numgot < RAREST_FIRST_CUTOFF)
	{
		unsigned int count_pieces = this->numpieces;
		while(count_pieces > 0)
		{
			count_pieces--;
			unsigned int random_int = (unsigned int)(numpieces * rand()/(RAND_MAX + 1.0));
			/* First check if this piece has been requested already or not */
			if(this->pieces_in_progress.find(random_int) != this->pieces_in_progress.end())
			{
				continue;
			}
			if(other_side_has(random_int, other_have))
			{
				/* I should not have this piece */
				if(other_side_has(random_int, this->local_bitfield))
				{
					continue;
				}
				else
					return random_int;
			}
		}
	}
	/* else choose the rarest piece */
	else
	{
		for(unsigned int i = 1; i < this->interests.size() ; i++)
		{
			vector<unsigned int>* temp_pieces = this->interests.at(i);
			for(unsigned int j = 0; j < temp_pieces->size(); j++)
			{
				/* first check if this piece has been requested already*/
				if(this->pieces_in_progress.find(temp_pieces->at(j)) != this->pieces_in_progress.end())
				{
					continue;
				}

				if(other_side_has(temp_pieces->at(j), other_have))
				{

					/* Choose another piece if I already have this one */
					if(other_side_has(temp_pieces->at(j), this->local_bitfield))
					{
						continue;
					}
					else
					{
						return temp_pieces->at(j);
					}
				}
			}
		}
	}

	return NO_NEXT_PIECE;
}

/* function to transfer a piece_number from one list to another (for piece_picker) */
void bt_peer::shift_over(unsigned int piece_number, vector<unsigned int> *old_list, vector<unsigned int> *new_list)
{
	unsigned int p = this->pos_in_interests[piece_number];


	unsigned int last_old = old_list->at(old_list->size() - 1);
	if(old_list->empty())
	{
		assert(0);
	}

	/* swap the last element with piece_number*/
	old_list->at(p) = last_old;
	this->pos_in_interests[last_old] = p;
	old_list->pop_back();
	
	/* get a random position in newList*/
	unsigned int newp = (unsigned int)((new_list->size() + 1)*rand()/(RAND_MAX + 1.0));
	
	if(newp >= new_list->size())
	{
		newp = new_list->size();
		this->pos_in_interests[piece_number] = newp;
		new_list->push_back(piece_number);
	}
	else
	{
		unsigned int old_ele = new_list->at(newp);
		this->pos_in_interests[old_ele] = new_list->size();
		new_list->push_back(old_ele);
		new_list->at(newp) = piece_number;
		this->pos_in_interests[piece_number] = newp;
	}
}


/* this function transfers a request for a piece to another downloader as the current one
   got either disconnected or choked in the middle of a a piece transfer*/
void bt_peer::transfer_request(bt_peer* other_side, unsigned int piece_number)
{

	/* no need to do anything in this case */
	if(piece_number == NO_NEXT_PIECE)
	{
		return ;
	}

	stdext::hash_map<bt_peer*, unsigned char,HashBtPeer, HashBtPeerCompare>::iterator dl_it;
	for(dl_it = this->dl_choked.begin(); dl_it != this->dl_choked.end(); dl_it++)
	{
		bt_peer* curr = dl_it->first;
		/* this downloader is not interested => maybe thinks that
			no more pieces are to be downloaded*/
		/*if(this->dl_interested[curr] == NOT_INTERESTED && this->dl_choked[curr] == CHOKED)*/
		if(this->dl_interested[curr] == NOT_INTERESTED)
		{
			/* if the other side of this downloader has this piece, make it interested */
			if(other_side_has(piece_number, curr->local_bitfield))
			{
				/*if(this->dl_choked[curr] == UNCHOKED)
				{
					log_msg("TRQ at %u in %u from %u to unchoked %u\n", g_tick,this->peer_id, other_side->peer_id, curr->peer_id);
				}
				else
				{
										
				}*/
				send_interested(curr);
			}
		}
	}
}


/* this function removes a uploader from the peer's state
 * state that contains uploader : 1)choker_connections 
 */
void bt_peer::remove_uploader(bt_peer* other_peer)
{
	/* remove the uploader from the choker's list of connections*/
	vector<bt_peer*>::iterator ccit;
	for(ccit = this->choker_connections.begin(); ccit != this->choker_connections.end(); ccit++)
	{
		if(*ccit == other_peer)
		{
			this->choker_connections.erase(ccit);
			break;
		}
	}
}

/*****************************************************************************
*-------------------downloader--------------------------------------------*
*****************************************************************************/


/* starts the functioning of the downloader*/
void bt_peer::download_startup(bt_peer* other_side)
{
	assert(other_side);

	/* since I have just got the bitfield => update the state for the picker also*/
	
	for(unsigned int i = 0; i < this->numpieces; i++)
	{	

		if(other_side_has(i, other_side->local_bitfield))
		{
			/* update state for the picker*/
			unsigned int numint = this->numinterests[i];
			if(numint == this->interests.size() - 1)
			{
				vector<unsigned int>* new_list = new vector<unsigned int>;
				this->interests.push_back(new_list);
			}

			this->numinterests[i] += 1;
			this->shift_over(i, this->interests[numint],this->interests[numint + 1]);

		}
	}
	if(!(this->done))
	{
		/* I am not complete */
		if(has_something(this->local_bitfield,other_side->local_bitfield))
		{
			/* send interested*/
			/*//DEBUG_MSG("Peer %u wants something from %u\n", this->peer->peer_id, this->other_side->peer->peer_id);*/
			send_interested(other_side);
		}
	}
}

/* sends an 'interested' message to the other side*/
void bt_peer::send_interested(bt_peer* other_side)
{

/*	if(other_side->peer_id == 1)
	{
		log_msg("IN in %u interested in seed at %u\n", this->peer_id, g_tick);
	}*/


	/* call the other side's got_interested function*/
	other_side->got_interested(this);
	unsigned char previous_int = this->dl_interested[other_side];
	this->dl_interested[other_side] = INTERESTED;

	if(this->dl_choked[other_side] == UNCHOKED && previous_int == NOT_INTERESTED)
	{
		/* Since I was unchoked and now I am interested also => start data transfer*/
		add_data_transfer(other_side, this);
		request_more(other_side);
	}
}

/* sends a 'not interested' message to the other side*/
void bt_peer::send_not_interested(bt_peer* other_side)
{
/*	if(other_side->peer_id == 1)
	{
		log_msg("NI in %u not interested in seed at %u\n", this->peer_id, g_tick);
	}*/


	if(this->dl_interested[other_side] == INTERESTED && this->dl_choked[other_side] == UNCHOKED)
	{
		/* since I was interested and also unchoked => transfer was going on */
		/* stop the transfer*/
		this->curr_piece[other_side] = NO_NEXT_PIECE;
		this->curr_piece_downloaded[other_side] = 0;
		remove_data_transfer(other_side, this);
	}
	this->dl_interested[other_side] = NOT_INTERESTED;

	/* call the other side's got_not_interested function*/
	other_side->got_not_interested(this);
}

/* sends another request after getting a piece*/
void bt_peer::request_more(bt_peer* other_side)
{
	/* request another piece => assume that the data_transfer is already on */
	/* since the request is not actually to be sent to the other side -
	   just update the info stored locally about which piece is next */


	/* first see if the entry for the other side exists in curr_piece*/
	if(this->curr_piece.find(other_side) != this->curr_piece.end())
	{
		/* there is some entry for this peer in curr_piece*/
		unsigned int temp_piece = this->curr_piece[other_side];
		if(temp_piece != NO_NEXT_PIECE)
		{
			/* another transfer got added in the middle => no need to override it*/
			assert(this->pieces_in_progress.find(temp_piece) != this->pieces_in_progress.end());
			return;
		}
	}

	this->curr_piece[other_side] = this->piece_picker(other_side->local_bitfield);
	this->curr_piece_downloaded[other_side] = 0;

	if(this->dl_choked[other_side] == UNCHOKED)
	{
		if(this->curr_piece[other_side] == NO_NEXT_PIECE)
		{
			/* all undownloaded pieces are currently in progress so nothing for you currently*/
			/* OR maybe the other side doesn;t have any new piece to offer */

			send_not_interested(other_side);
		}
		else
		{
			/* also add this piece to the pieces being requested currently*/
			this->pieces_in_progress[this->curr_piece[other_side] ] = true;
		}
	}
}

/* function to receive bytes from the network 
 * params: 1) number of bytes transferred */
void bt_peer::got_data(bt_peer* other_side, unsigned int numbytes)
{
	/* CHECK THIS*/
	/* hacky check to see if I shudn't be getting this data */
	if(this->curr_piece[other_side] == NO_NEXT_PIECE)
	{
		//DEBUG_MSG("There should be no data\n");
		return;
	}


	this->curr_piece_downloaded[other_side] += numbytes;
	if(this->curr_piece_downloaded[other_side] >= PIECE_LEN)
	{
		/* now this piece download is complete*/
		got_piece(other_side, this->curr_piece[other_side]);
	}
}

/* receives a complete piece */
void bt_peer::got_piece(bt_peer* other_side, unsigned int piece_number)
{
	/* update whatever state needs to be updated*/
	add_to_bitfield(this->local_bitfield,piece_number);
	stdext::hash_map<unsigned int, bool>::iterator pipit;
	pipit = this->pieces_in_progress.find(piece_number);
	if(pipit != this->pieces_in_progress.end())
	{
		this->pieces_in_progress.erase(pipit);
	}
	else
	{
		fprintf(stderr,"Warning(GOT_PIECE) : This piece(%u) should have been in progress !\n",piece_number);
	}
	this->numgot++;
	this->curr_piece[other_side] = NO_NEXT_PIECE;
	this->curr_piece_downloaded[other_side] = 0;


	log_msg("Piece: %u: %u: %u: %u (Total = %u out of %u)\n", g_tick, this->peer_id, other_side->peer_id, piece_number, this->numgot, this->numpieces);
	/* send have messages to all downloaders */
	stdext::hash_map<bt_peer*, unsigned char,HashBtPeer, HashBtPeerCompare>::iterator ul_it;
	for(ul_it = this->ul_choked.begin(); ul_it != this->ul_choked.end(); ul_it++)
	{
		bt_peer* curr = ul_it->first;
		send_have(curr, piece_number);
	}

	/* update rate information */
	download_update_rate(other_side, PIECE_LEN);
	/* also call the uploader's update rate as it doesn't know when to call it*/
	other_side->upload_update_rate(this, PIECE_LEN);

	if(i_am_complete(this->local_bitfield))
	{
		this->curr_piece[other_side] = NO_NEXT_PIECE;
		this->curr_piece_downloaded[other_side] = 0;

		this->done = true;
		/* close one connection  => since a transfer was going on remove it*/

		send_not_interested(other_side);

		/* schedule a leave event for the future*/
		/* nodes do not leave*/
		
		event* leave_event = event_alloc();
		leave_event->fire_time = this->grace_seed_time + g_tick;
		leave_event->peer = this;
		leave_event->type = E_LEAVE;
		event_insert(leave_event);
	}
	else
	{
		if(this->dl_choked[other_side] == UNCHOKED)
		{
			/* request more pieces */
			request_more(other_side);
		}
	}
}

/* receive choke message */
void bt_peer::got_choke(bt_peer* other_side)
{
	if(this->is_active_peer == false)
	{
		return;
	}

	/* indicated which piece's tranfer was aborted*/
	unsigned int gone_piece = NO_NEXT_PIECE;
	/* update state */
	if(this->dl_interested[other_side] == INTERESTED && this->dl_choked[other_side] == UNCHOKED)
	{
		/* This means that a transfer was going on => remove the transfer */
		stdext::hash_map<unsigned int, bool>::iterator pipit;
		pipit = this->pieces_in_progress.find(this->curr_piece[other_side]);
		if(pipit != this->pieces_in_progress.end())
		{
			gone_piece = this->curr_piece[other_side];
			this->pieces_in_progress.erase(pipit);
			this->curr_piece[other_side] = NO_NEXT_PIECE;

			/*unsigned int lsize = this->curr_piece_downloaded[other_side]; 
			if(lsize > 0)
			{
				log_msg("LSUBP: %u %u\n", this->peer_id, lsize);
			}*/
			this->curr_piece_downloaded[other_side] = 0;

		}
		else
		{
			/*fprintf(stderr,"Warning(GOT_CHOKE) %u : This piece(%u) should have been in progress %u %u!\n",g_tick,this->curr_piece[other_side], this->peer_id, this->numgot);
			log_msg("Warning(GOT_CHOKE) %u : This piece(%u) should have been in progress %u %u %u!\n",g_tick,this->curr_piece[other_side], this->peer_id, this->numgot, other_side->peer_id);*/
		}

		remove_data_transfer(other_side, this);
	}
	this->dl_choked[other_side] = CHOKED;
	this->transfer_request(other_side, gone_piece);
}

/* receive unchoke */
void bt_peer::got_unchoke(bt_peer* other_side)
{
	if(this->is_active_peer == false)
	{
		return;
	}


	/* update state */
	unsigned char previous_choke = this->dl_choked[other_side];
	this->dl_choked[other_side] = UNCHOKED;

	if(this->dl_interested[other_side] == INTERESTED && previous_choke == CHOKED)
	{
		/* I am interested and unchoked now => Can get data */
		/* CHECK what is the correct value of time */

		add_data_transfer(other_side, this);
		request_more(other_side);
	}
}

/* receive a have message */
void bt_peer::got_have(bt_peer* other_side, unsigned int piece_number)
{

	if(this->is_active_peer == false)
		return;
	if(this->done == true)
		return;

	/* update state */
	
	unsigned int numint = this->numinterests[piece_number];
	if(numint == this->interests.size() - 1)
	{
		vector<unsigned int>* new_list = new vector<unsigned int>;
		this->interests.push_back(new_list);
	}
	
	//fprintf(stderr,"Number of lists: %u, Size of first: %u, Second: %u\n", this->interests.size(), this->interests[0]->size(), this->interests[1]->size());


	this->numinterests[piece_number] += 1;
	this->shift_over(piece_number,this->interests[numint],this->interests[numint + 1]);

	/* check if need this piece => if yes, then send interested if I am not interested currently */
	if(!other_side_has(piece_number, this->local_bitfield))
	{
		/* I want this piece */
		if(this->dl_interested[other_side] == NOT_INTERESTED)
			send_interested(other_side);
	}
}

/*rate updater function => param: 1) amount of data received since last update */
void bt_peer::download_update_rate(bt_peer* other_side, unsigned int amount)
{
	unsigned int curr_time = g_tick;

	/* find what is the current time */
	this->dl_rate[other_side] = (this->dl_rate[other_side] * (this->dl_last_time[other_side] - this->dl_ratesince_time[other_side]) + amount)/(curr_time - this->dl_ratesince_time[other_side]);
	this->dl_last_time[other_side] = curr_time;
	if (this->dl_ratesince_time[other_side] < (curr_time - MAX_RATE_PERIOD))
		this->dl_ratesince_time[other_side] = curr_time - MAX_RATE_PERIOD;
}


/* this function cleans up state for this downloader
 * may be called when: 1) this DOWNLOAD CONNECTION got closed by other side
                       2) this PEER decides to leave the network */
void bt_peer::download_cleanup_state(bt_peer* other_side, unsigned char clean_choke)
{
	/* remove this downloader from all download state*/
	stdext::hash_map<bt_peer*, unsigned char,HashBtPeer, HashBtPeerCompare>::iterator dl_it;

	if(clean_choke == CLEAN_CHOKE)
	{
		dl_it = this->dl_choked.find(other_side);
		if(dl_it != this->dl_choked.end())
		{
			/* found the entry */
			this->dl_choked.erase(dl_it);
		}
		else
		{
			fprintf(stderr,"Warning: Should have found the downloader entry in dl_choked\n");
		}
	}
	dl_it = this->dl_interested.find(other_side);
	if(dl_it != this->dl_interested.end())
	{
		/* found the entry */
		this->dl_interested.erase(dl_it);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_interested\n");
	}

	stdext::hash_map<bt_peer*, unsigned int,HashBtPeer, HashBtPeerCompare>::iterator dl_it_rate;
	dl_it_rate = this->dl_rate.find(other_side);
	if(dl_it_rate != this->dl_rate.end())
	{
		/* found the entry */
		this->dl_rate.erase(dl_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_rate\n");
	}

	dl_it_rate = this->dl_ratesince_time.find(other_side);
	if(dl_it_rate != this->dl_ratesince_time.end())
	{
		/* found the entry */
		this->dl_ratesince_time.erase(dl_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_ratesince_time\n");
	}

	dl_it_rate = this->dl_last_time.find(other_side);
	if(dl_it_rate != this->dl_last_time.end())
	{
		/* found the entry */
		this->dl_last_time.erase(dl_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_last_time\n");
	}

	stdext::hash_map<bt_peer*, unsigned int,HashBtPeer, HashBtPeerCompare>::iterator dl_piece;
	dl_piece = this->curr_piece.find(other_side);
	if(dl_piece != this->curr_piece.end())
	{
		/* found the entry */
		this->curr_piece.erase(dl_piece);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_piece\n");
	}

	dl_piece = this->curr_piece_downloaded.find(other_side);
	if(dl_piece != this->curr_piece_downloaded.end())
	{
		/* found the entry */
		this->curr_piece_downloaded.erase(dl_piece);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the downloader entry in dl_piece_downloaded\n");
	}

}

/* this function is called when the other side has closed this connection
 * things to be done here:
 * 1) transfer requests and close this transfer 
 * 2) cleanup state							*/
void bt_peer::download_close_connection(bt_peer* other_side)
{
	this->numconnections--;

	/*  now transfer the request --- 
		indicates which piece's tranfer was aborted*/
	unsigned int gone_piece = NO_NEXT_PIECE;

	if(this->dl_interested[other_side] == INTERESTED && this->dl_choked[other_side] == UNCHOKED)
	{
		/* This means that a transfer was going on => remove the transfer */
		stdext::hash_map<unsigned int, bool>::iterator pipit;
		pipit = this->pieces_in_progress.find(this->curr_piece[other_side]);
		if(pipit != this->pieces_in_progress.end())
		{
			gone_piece = this->curr_piece[other_side];
			this->pieces_in_progress.erase(pipit);
			this->curr_piece[other_side] = NO_NEXT_PIECE;
			this->curr_piece_downloaded[other_side] = 0;
		}
		else
		{
			fprintf(stderr,"Warning(CLOSE_CONN) : This piece(%u) should have been in progress !\n",this->curr_piece[other_side]);
		}

		remove_data_transfer(other_side, this);
	}

	this->dl_choked[other_side] = CHOKED;

	this->transfer_request(other_side, gone_piece);

	this->download_cleanup_state(other_side, CLEAN_CHOKE);

}

void bt_peer::check_min_connections()
{
	/* now check if the num connections has gone down below the min level*/
	if(this->numconnections < MIN_PEERS)
	{
		/* rerequest => code copied because I dont want to add another rerequest event*/
		if(this->is_active_peer)
		{
			/* Get the tracker pointer*/
			bt_tracker* mytracker = get_tracker(this);

			/* get list of peers from client */
			vector<bt_peer*> *got_list = mytracker->announce(this);

			vector<bt_peer*>::iterator peer_it;

			/* traverse the list of peers received*/
			for(peer_it = got_list->begin(); peer_it != got_list->end(); peer_it++)
			{
				if(this->numconnections < MAX_SELF_INITIATE)
				{
					int status = this->start_connection(*peer_it);
					if(status == NEW_PEER)
						this->numconnections++;
				}
				else
				{
					/* Add this peer to spare connections */
					break;
				}
			}

			log_msg("Imm Rerequest: %u: %u: %u: %u: %u: %u: %u\n", g_tick, this->peer_id,got_list->size(), this->numconnections, this->numgot, this->pieces_in_progress.size(), this->num_downloads_curr()) ;

		}

	}

}



/*****************************************************************************
******************************************************************************/

/* receive a request */
void bt_peer::got_request(bt_peer* other_side)
{
	if(this->ul_interested[other_side] == NOT_INTERESTED)
	{
		fprintf(stderr,"Protocol Error : got request before an interested message\n");
	}
	else
	{
		/* CHECK HOW TO SEND THE PIECE */
	}
}

/* receive a cancel */
void bt_peer::got_cancel(bt_peer* other_side)
{

}

/* choke the other guy*/
void bt_peer::send_choke(bt_peer* other_side)
{
	this->ul_choked[other_side] = CHOKED;

	other_side->got_choke(this);
}

/* unchoke the other guy */
void bt_peer::send_unchoke(bt_peer* other_side)
{
	this->ul_choked[other_side] = UNCHOKED;
	
	other_side->got_unchoke(this);
}

/* receive an 'interested' message */
void bt_peer::got_interested(bt_peer* other_side)
{
	/* update state */
	this->ul_interested[other_side] = INTERESTED;

	/* call the rechoke mechanism if the other side is unchoked*/
	if(this->ul_choked[other_side] == UNCHOKED)
	{
		this->rechoke();
	}
}

/* receive not_interested */
void bt_peer::got_not_interested(bt_peer* other_side)
{
	/* update state */
	this->ul_interested[other_side] = NOT_INTERESTED;
	
	if(this->ul_choked[other_side] == UNCHOKED)
	{
		this->rechoke();
	}
}

/* sends a 'have' message to the other side */
void bt_peer::send_have(bt_peer* other_side, unsigned int piece_number)
{
	/* call other side's got_have fucntion*/
	other_side->got_have(this, piece_number);
}

/*rate updater function => param: 1) amount of data received since last update */
void bt_peer::upload_update_rate(bt_peer* other_side, unsigned int amount)
{
	unsigned int curr_time = g_tick;

	/* find what is the current time */
	this->ul_rate[other_side] = (this->ul_rate[other_side] * (this->ul_last_time[other_side] - this->ul_ratesince_time[other_side]) + amount)/(curr_time - this->ul_ratesince_time[other_side]);
	this->ul_last_time[other_side] = curr_time;
	if (this->ul_ratesince_time[other_side] < (curr_time - MAX_RATE_PERIOD))
		this->ul_ratesince_time[other_side] = curr_time - MAX_RATE_PERIOD;
}

/* this function cleans up state for this uploader 
 * may be called when: 1) this UPLOAD CONNECTION got closed by other side
                       2) this PEER decides to leave the network */
void bt_peer::upload_cleanup_state(bt_peer* other_side)
{
	/* remove this uploader from all upload state*/
	stdext::hash_map<bt_peer*, unsigned char,HashBtPeer, HashBtPeerCompare>::iterator ul_it;
	ul_it = this->ul_choked.find(other_side);
	if(ul_it != this->ul_choked.end())
	{
		/* found the entry */
		this->ul_choked.erase(ul_it);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the uploader entry in ul_choked\n");
	}
	ul_it = this->ul_interested.find(other_side);
	if(ul_it != this->ul_interested.end())
	{
		/* found the entry */
		this->ul_interested.erase(ul_it);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the uploader entry in ul_interested\n");
	}

	stdext::hash_map<bt_peer*, unsigned int,HashBtPeer, HashBtPeerCompare>::iterator ul_it_rate;
	ul_it_rate = this->ul_rate.find(other_side);
	if(ul_it_rate != this->ul_rate.end())
	{
		/* found the entry */
		this->ul_rate.erase(ul_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the uploader entry in ul_rate\n");
	}

	ul_it_rate = this->ul_ratesince_time.find(other_side);
	if(ul_it_rate != this->ul_ratesince_time.end())
	{
		/* found the entry */
		this->ul_ratesince_time.erase(ul_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the uploader entry in ul_ratesince_time\n");
	}

	ul_it_rate = this->ul_last_time.find(other_side);
	if(ul_it_rate != this->ul_last_time.end())
	{
		/* found the entry */
		this->ul_last_time.erase(ul_it_rate);
	}
	else
	{
		fprintf(stderr,"Warning: Should have found the uploader entry in ul_last_time\n");
	}
}

/* this function is called when the other side has closed this connection
 * things to be done here:
 * 1) inform my parent peer that this connection is gone
 * 2) cleanup state							*/
void bt_peer::upload_close_connection(bt_peer* other_side)
{
	/* also remove this uploader from the peer's state*/
	this->remove_uploader(other_side);

	/* now cleanup state */
	this->upload_cleanup_state(other_side);
}
