#include <assert.h>
#include <vector>
#include <algorithm>
#include <map>

using namespace std;

#include "bt_subtracker.h"
#include "bt_peer.h"
#include "topology.h"
#include "util.h"


bt_subtracker::bt_subtracker()
{
	tracker_id = 0;
	_tracker = NULL;
	_external_track_clients = new vector<bt_peer*>();
}

void bt_subtracker::set_tracker(bt_tracker *tracker)
{
	assert(tracker);
	_tracker = tracker;
}

vector<bt_peer*>* bt_subtracker::announce(bt_peer* caller)
{
	assert(caller);
	assert(_tracker);

	/* first check if the connecting peer is already present in list*/
	map<bt_peer*, bool>::iterator check_peer_it;
	check_peer_it = map_clients.find(caller);
	if(check_peer_it == map_clients.end()) {
		/* Add peer if not already present */
		_local_clients.push_back(caller);	
		map_clients[caller] = true;
	}

	vector<bt_peer*>* tracker_clients = _tracker->announce(caller);

	unsigned int num_local_clients = _local_clients.size();
	unsigned int num_track_clients_needed;
	unsigned int num_local_clients_needed;


	/* added by Ruchir(Jun 29'05) to support per AS number of external connections*/

	/* now check if the tracker limits the connections on a per AS basis
				or a per node basis */
	if(get_tracker_type() == SUB_LIMIT_TRACKER)
	{
		/* check how many external clients I have*/

		/* first remove the dead peers from the external list*/
		vector<bt_peer*> * new_external = new vector<bt_peer*>();	

		vector<bt_peer*>::iterator peer_it = _external_track_clients->begin();
		for(; peer_it != _external_track_clients->end(); peer_it++)
		{
			if((*peer_it)->is_active_peer == true)
			{
				new_external->push_back(*peer_it);
			}
		}
		
		_external_track_clients->clear();
		delete _external_track_clients;
		
		_external_track_clients = new_external;

		unsigned int num_track_clients = _external_track_clients->size();
		if(num_track_clients < get_num_external_peers())
		{
			num_track_clients_needed = get_num_external_peers() - num_track_clients;

			/* add some more peers to the external track clients list*/
			vector<bt_peer*>::iterator track_it = tracker_clients->begin();
			unsigned int num_track_clients_used = 0;

			for (; track_it != tracker_clients->end() && num_track_clients_used < num_track_clients_needed;	track_it++)
			{
				bt_peer *neighbor = *track_it;
				vector<bt_peer*>::iterator it;
				for (it = _local_clients.begin(); it != _local_clients.end(); it++) 
				{
					bt_peer *local_neighbor = *it;
					if (local_neighbor == neighbor) 
					{
						break;
					}
				}
				/* don't add if the external neighbor is actually local */
				if (it == _local_clients.end()) {
					num_track_clients_used++;
					_external_track_clients->push_back(neighbor);
				}
			}

		}

		/* now make the vector of clients to be returned*/
		num_track_clients = _external_track_clients->size();
		num_local_clients_needed = MAX_SELF_INITIATE - num_track_clients;

		clients.clear();
		peer_it = _external_track_clients->begin();
		for(; peer_it != _external_track_clients->end(); peer_it++)
		{
			unsigned int rand_int = (unsigned int)(100 * rand()/(RAND_MAX + 1.0));
			if(rand_int % 2 == 0)
			{
				bt_peer* neighbor = *peer_it;
				clients.push_back(neighbor);
			}
		}
		

		random_shuffle(_local_clients.begin(), _local_clients.end());

		peer_it = _local_clients.begin();
		for (peer_it = _local_clients.begin();	
					peer_it != _local_clients.end() && num_local_clients_needed > 0;
						peer_it++)
		{
			clients.push_back(*peer_it);
			num_local_clients_needed--;
		}
	
		random_shuffle(clients.begin(), clients.end());
		return &clients;

	}
	else
	{
		/* per node connection limiter*/

		/* determine the number of local / external neighbors necessary */

		if (num_local_clients > MAX_SELF_INITIATE - get_num_external_peers()) {
			num_track_clients_needed = get_num_external_peers();
			num_local_clients_needed = MAX_SELF_INITIATE -
				get_num_external_peers();
		} else {
			num_track_clients_needed = MAX_SELF_INITIATE -
				num_local_clients;
			num_local_clients_needed = num_local_clients;
		}

		/* add the external neighbors */

		vector<bt_peer*>::iterator track_it = tracker_clients->begin();
		unsigned int num_track_clients_used = 0;

		clients.clear();
		for (; track_it != tracker_clients->end() &&
				num_track_clients_used < num_track_clients_needed;
				track_it++) {
			bt_peer *neighbor = *track_it;
			vector<bt_peer*>::iterator it;
			for (it = _local_clients.begin(); it != _local_clients.end();
					it++) {
				bt_peer *local_neighbor = *it;
				if (local_neighbor == neighbor) {
					break;
				}
			}
			/* don't add if the external neighbor is actually local */
			if (it == _local_clients.end()) {
				num_track_clients_used++;
				clients.push_back(neighbor);
			}
		}

		/* add the local neighbors */

		random_shuffle(_local_clients.begin(), _local_clients.end());

		vector<bt_peer*>::iterator local_it;
		unsigned int num_local_clients_used = 0;

		for (local_it = _local_clients.begin();
				local_it != _local_clients.end() &&
				num_local_clients_used < num_local_clients_needed;
				local_it++, num_local_clients_used++) {
			clients.push_back(*local_it);
		}

		random_shuffle(clients.begin(), clients.end());
		return &clients;
	}
}

void bt_subtracker::peer_leaving(bt_peer* leaving_peer)
{
	assert(leaving_peer);
	assert(_tracker);
	_tracker->peer_leaving(leaving_peer);
	map<bt_peer*, bool>::iterator pit;
	pit = this->map_clients.find(leaving_peer);
	if(pit != this->map_clients.end())
	{
		/* peer found in records*/
		this->map_clients.erase(pit);
	} else {
		DEBUG_MSG("This peer should have been in the subtracker's map\n");
		return;
	}

	vector<bt_peer*>::iterator lpit;
	for(lpit = _local_clients.begin();
			lpit != _local_clients.end();
			lpit++)
	{
		/* check for the peer*/
		if(*lpit == leaving_peer)
		{
			/* remove this peer*/
			_local_clients.erase(lpit);
			return;
		}
	}
	BUG();
}
