#ifndef __MEMORY_POOL_H__
#define __MEMORY_POOL_H__

#include <stdlib.h>

struct memory_pool {
	void *pool_objects;			/* objects stored in pool */
	unsigned int num_objects;		/* number of objects */
	size_t obj_size;			/* size of an object */
	unsigned int *used_bitmap;		/* bitmap of used objects
						   0 = unused
						   1 = used */
	unsigned int num_words;			/* number of words in bitmap */
	unsigned int next_word_offset;		/* next word to search for 
						   free object bit in map */
};

/**
 * Initializes the memory pool
 * @param pool		the memory pool
 * @param num_objects	the number of objects to allocate in the pool
 * @param obj_size	size of an object
 */
void init_memory_pool(struct memory_pool *pool,
		unsigned int num_objects,
		size_t obj_size);

/**
 * Obtains an object from a memory pool.  Marks it as used.
 * @param pool		the memory pool
 * @return a pointer to the object, or NULL if out of available objects
 */
void *pool_object_alloc(struct memory_pool *pool);

/**
 * Frees an object.  Marks it as unused.
 * @param pool		the memory pool
 * @param obj		the object to be freed
 */
void pool_object_free(struct memory_pool *pool, void *obj);

#endif /* __MEMORY_POOL_H__ */
